import sys
import tty
import termios
from typing import List, Optional
from ctube.colors import Color, color


class Prompt:
    def __init__(self, char: str = "❯", col: Color = Color.GREEN):

        self.char = char
        self.col = col

    def get_input(self):
        return input(f"{color(self.char, self.col)} ")


class Buffer:
    def __init__(self, infinite_scroll: bool = True):
        self.infinite_scroll = infinite_scroll
        self._index = 0
        self._buffer: List[str] = []

    def get(self) -> Optional[str]:
        return self._buffer[self._index]

    def next(self) -> None:
        if self._index == len(self._buffer) - 1:
            if self.infinite_scroll:
                self._index = 0
        else:
            self._index += 1

    def prev(self) -> None:
        if self._index == 0:
            if self.infinite_scroll:
                self._index = len(self._buffer) - 1
        else:
            self._index -= 1

    def add(self, string: str) -> None:
        self._buffer.append(string)


def get_fchar_unix() -> str:
    charlist: List[str] = []

    for i in range(3):
        charlist.append(get_char_unix())

        # chr(27): '\x1b'
        # chr(91): '['

        if charlist[i] not in [chr(27), chr(91)]:
            break

        if len(charlist) > 1:
            if charlist == [chr(27), chr(27)]:
                break

    if len(charlist) == 3:
        if charlist[2] == "A":
            return "UP"
        if charlist[2] == "B":
            return "DOWN"
        if charlist[2] == "C":
            return "RIGHT"
        if charlist[2] == "D":
            return "LEFT"

    if len(charlist) == 2:
        if charlist == [chr(27), chr(27)]:
            return chr(27)

    if len(charlist) == 1:
        return charlist[0]

    return ""


def get_char_unix() -> str:
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch



import re
import requests
from typing import List , Callable, Dict
from urllib.error import URLError
from httpx import ReadTimeout, ConnectTimeout, ConnectError
from innertube.errors import RequestError
from ctube.containers import Album
from ctube.errors import InvalidIndexSyntax
from ctube.parser import parse_indexes
from ctube.colors import Color
from ctube.printers import write


def filter_albums_by_indexes(albums: List[Album], user_input: str) -> List[Album]:
    selected_indexes = parse_indexes(user_input)
    if not selected_indexes:
        return albums

    elif isinstance(selected_indexes, slice):
        selected_albums = albums[selected_indexes]
        if not selected_albums:
            if selected_indexes.start == selected_indexes.stop - 1:
                raise InvalidIndexSyntax(f"Invalid index: {selected_indexes.start}")
            else:
                raise InvalidIndexSyntax(f"Invalid slice: {selected_indexes.start}:{selected_indexes.stop}")
        else:
            return selected_albums
    else:
        incorrect_indexes: List[int] = []
        for index in selected_indexes:
            if index >= len(albums):
                incorrect_indexes.append(index)
        if incorrect_indexes:
            str_incorrect_indexes = ', '.join(map(str, incorrect_indexes))
            if len(incorrect_indexes) > 1:
                raise InvalidIndexSyntax(f"Invalid indexes: {str_incorrect_indexes}")
            else:
                raise InvalidIndexSyntax(f"Invalid index: {str_incorrect_indexes}")
        else:
            return [albums[index] for index in selected_indexes]


def filter_albums_by_regex(albums: List[Album], pattern: str) -> Dict[int, Album]:
    filtered_albums: Dict[int, Album] = {}
    for i in range(len(albums)):
        res = re.search(string=albums[i].title, pattern=pattern)
        if res:
            filtered_albums[i] = albums[i]
    return filtered_albums


def handle_connection_errors(func: Callable) -> Callable:
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestError:
            write("Invalid request", Color.RED)
        except (ConnectTimeout, ReadTimeout):
            write("A timeout error occurred. Try again.", Color.RED)
        except (ConnectError, URLError):
            write("No internet connection", Color.RED)
    return inner


def connected_to_internet(url: str = 'http://www.google.com/', timeout: int = 5) -> bool:
    try:
        _ = requests.head(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        return False



import os
import sys
import shutil
import eyed3
from pydub import AudioSegment
from ctube.containers import Song


def on_progress_callback(
        song: Song,
        filesize: int, 
        bytes_received: int, 
) -> None:
    columns = shutil.get_terminal_size().columns
    max_width = int(columns * 0.40)
    filled = int(round(max_width * bytes_received / float(filesize)))
    remaining = max_width - filled
    progress_bar = "#" * filled + "-" * remaining
    percent = round(100.0 * bytes_received / float(filesize), 1)

    distance_from_bar = columns - (max_width + 9)  # len bar + percentage len
    title = f":: {song.title} "

    if len(title) > distance_from_bar:
        title = f"{title[:distance_from_bar - 4]}... "
    else:
        title = f"{title}{' ' * (distance_from_bar - len(title))}"

    text = f"{title}[{progress_bar}] {percent}%\r"

    sys.stdout.write(text)
    sys.stdout.flush()


def on_complete_callback(song: Song) -> None:
    output = f"{os.path.splitext(song.filepath)[0]}.mp3"
    mp4 = AudioSegment.from_file(song.filepath, "mp4")
    mp4.export(output, format="mp3")
    set_metadata(filepath=output, song=song)
    os.remove(song.filepath)


def set_metadata(filepath: str, song: Song) -> None:
    audio = eyed3.load(filepath)
    album = song.album
    if audio and audio.tag:
        tag = audio.tag
        # Song tag
        tag.title = song.title
        tag.artist = song.artist
        tag.track_num = song.track_num
        tag.images.set(3, song.image_data, "image/jpeg", u"cover")

        # Album tag
        # eyed3.tag supported album_type values:
        # lp, ep, compilation, live, various, demo, sigle
        album_type = album.album_type.lower()
        if album_type != "album":
            try:
                tag.album_type = album_type 
            except ValueError:
                pass  # TODO: log error
        tag.album = album.title
        tag.release_year = album.release_year
        tag.save()



import shutil
from typing import List, Dict
import ctube
from ctube.cmds import Command
from ctube.colors import color, Color
from ctube.containers import Album


def print_header() -> None:
    write("░█▀▀░▀█▀░█░█░█▀▄░█▀▀░", Color.GREEN)
    write("░█░░░░█░░█░█░█▀▄░█▀▀░", Color.GREEN)
    write("░▀▀▀░░▀░░▀▀▀░▀▀░░▀▀▀░", Color.GREEN)
    print(f"\u2022 {color('version', Color.BOLD)}: {color(ctube.__version__, Color.BLUE)}")
    print("\u2022 source: https://github.com/g3nsy/ctube")
    print("\u2022 Type 'help' to list the available commands")


def print_help() -> None:
    write(":: Helper", Color.GREEN)
    for cmd in Command:
        print(f"\u2022 {color(cmd.value.name, Color.BOLD)}:")
        write(cmd.value.description, Color.GREEN)
        print()


def print_albums_list(albums: List[Album]) -> None:
    max_index_len = len(str(len(albums)))
    title_space = _get_title_space(max_index_len)
    for i, album in enumerate(albums):
        _print_album_title(i, album.title, title_space, max_index_len)


def print_albums_dict(albums: Dict[int, Album]) -> None:
    max_index_len = len(str(max(albums.keys())))
    title_space = _get_title_space(max_index_len)
    for i in albums:
        _print_album_title(i, albums[i].title, title_space, max_index_len)


def _get_title_space(max_index_len: int) -> int:
    terminal_columns = shutil.get_terminal_size().columns
    return terminal_columns - max_index_len - 3  # [, ], ' '

def _print_album_title(index: int, title: str, title_space: int, max_index_len: int):
    if len(title) > title_space:
        title = f"{title[:title_space - 3]}..."
    lb = color("[", Color.BLUE)
    rb = color("]", Color.BLUE)
    ci = color(str(index), Color.GREEN)
    print(f"{lb}{ci}{rb}{' ' * (1 + max_index_len - len(str(index)))}{title}")


def clear_screen() -> None:
    print("\033c", end="")


def write(string: str, col: Color = Color.WHITE):
    print(color(string, col=col))



from dataclasses import dataclass


@dataclass
class Album:
    title: str
    album_type: str
    release_year: int
    thumbnail_url: str
    playlist_id: str


@dataclass
class Song:
    title: str
    artist: str
    track_num: int
    image_data: bytes
    filepath: str
    album: Album



import sys
import socket
from typing import Optional
from signal import signal, SIGINT
from pytubefix import request
from ctube.app import App
from ctube.paths import MUSIC


def signal_handler(_: Optional[int] = None):
    print('\033[?25h', end="")
    sys.exit(0)


socket.setdefaulttimeout(3)
signal(SIGINT, lambda signum, _: signal_handler(signum))
request.default_range_size = request.default_range_size // 15


def main() -> None:
    app = App(
        output_path=MUSIC, 
        skip_existing=False,
        timeout=5,
        max_retries=3
    )
    app.main_loop()
    signal_handler()


if __name__ == "__main__":
    main()



import ctube
import sys
from requests.exceptions import RequestException
from typing import List, Optional, Tuple
from urllib import request
from urllib.error import URLError
from innertube.clients import InnerTube
from ctube.update import get_latest_version
from ctube.download import Downloader
from ctube.errors import InvalidIndexSyntax
from ctube.terminal import Prompt
from ctube.containers import Album
from ctube.colors import Color
from ctube.cmds import Command
from ctube.helpers import (
    filter_albums_by_indexes, 
    filter_albums_by_regex,
    handle_connection_errors,
    connected_to_internet
)
from ctube.parser import parse_user_input
from ctube.callbacks import (
    on_progress_callback, 
    on_complete_callback
)
from ctube.printers import (
    clear_screen,
    print_albums_list,
    print_albums_dict,
    print_header, 
    print_help, 
    write
)
from ctube.extractors import (
    extract_artist_id, 
    extract_albums
)


class App:
    def __init__(
            self, 
            output_path: str,
            skip_existing: bool = True,
            timeout: int = 5,
            max_retries: int = 3
    ):
        self.client = InnerTube("WEB_REMIX")
        self.prompt = Prompt()
        self.downloader = Downloader(
            output_path=output_path, 
            skip_existing=skip_existing,
            on_complete_callback=on_complete_callback,
            on_progress_callback=on_progress_callback,
            timeout=timeout,
            max_retries=max_retries
        )

        # last search
        self._albums: Optional[List[Album]] = None
        self._artist_name: Optional[str] = None

    def main_loop(self) -> None:
        clear_screen()
        print_header()
        try:
            latest_version = get_latest_version("ctube")
        except RequestException:
            pass
        else:
            if latest_version > ctube.__version__:
                print(f"=> New version available: {latest_version}")
                print(f"=> Close and run 'pip install -U ctube' to update.")
            else:
                print("=> ctube is up to date")
        while True:
            user_input = self.prompt.get_input().strip()

            if not user_input: 
                continue

            cmd_name, args = parse_user_input(user_input)

            try:
                cmd = Command.get_by_name(cmd_name)
            except KeyError:
                write(f"Invalid command: {cmd_name}", Color.RED)
            else:
                match cmd:
                    case Command.EXIT:
                        App._exit()
                    case Command.CLEAR:
                        clear_screen()
                    case Command.HELP:
                        print_help()
                    case Command.SEARCH:
                        self._search(args)
                    case Command.ID:
                        self._id(args)
                    case Command.DOWNLOAD:
                        self._download(args)
                    case Command.FILTER:
                        self._filter(args)

    @handle_connection_errors
    def _search(self, artist_name: str) -> None:
        if not artist_name:
            write("Missing argument: artist name", Color.RED)
        else:
            data = self.client.search(artist_name)
            try:
                artist_id = extract_artist_id(data)
            except (KeyError, TypeError, IndexError):
                write(f"Artist '{artist_name}' not found", Color.RED)
            else:
                self._id(artist_id)

    @handle_connection_errors
    def _id(self, artist_id: str) -> None:
        if not artist_id:
            write("Missing argument: artist id", Color.RED)
        else:
            data = self.client.browse(f"MPAD{artist_id}")
            try:
                artist_music_data = extract_albums(data)
            except (KeyError, TypeError, IndexError):
                write(f"Content not found", Color.RED)
            else:
                artist_music_data: Tuple[List[Album], str]
                self._albums, self._artist_name = artist_music_data
                self._albums.sort(key=lambda x: x.title)
                write(f"Collected music for {self._artist_name}", Color.GREEN)
                print_albums_list(self._albums)

    def _download(self, indexes: str):
        if not self._albums or not self._artist_name:
            write("You need to search for music first.", Color.RED)
            write("Use the search/id command", Color.RED)
        elif not indexes:
            write("Missing argument: indexes", Color.RED)
        elif not connected_to_internet():
            write("No internet connection", Color.RED)
        else:
            try:
                albums = filter_albums_by_indexes(self._albums, indexes)
            except InvalidIndexSyntax as error:
                write(str(error), Color.RED)
            else:
                print('\033[?25l', end="")
                if len(albums) == len(self._albums):
                    write(f"Selected items: ALL", Color.BLUE)
                else:
                    write(f"Selected items:", Color.BLUE)

                for album in albums:
                    write(f"\u2022 {album.title}", Color.BOLD)

                for album in albums:
                    write(f":: Downloading: {album.title}", Color.GREEN)

                    try:
                        response = request.urlopen(album.thumbnail_url)
                        image_data = response.read()
                    except URLError:
                        write(f"An error occurred while downloading {album.title} cover art", Color.RED)
                        write(f"Skipping {album.title}", Color.YELLOW)
                    else:
                        for song, error in self.downloader.download_album(
                                album=album, 
                                artist=self._artist_name,
                                image_data=image_data
                        ):
                            print()
                            if error:
                                write(f"An error occurred while downloading {song.title}", Color.RED)
                                write(f"Reason: {str(error)}", Color.RED)
                print('\033[?25h', end="")

    def _filter(self, pattern: str) -> None:
        if not self._albums or not self._artist_name:
            write("You need to search for music first.", Color.RED)
            write("Use the search/id command", Color.RED)
        elif not pattern:
            write("Missing argument: pattern", Color.RED)
        else:
            filtered_albums = filter_albums_by_regex(self._albums, pattern=pattern)
            if filtered_albums:
                print_albums_dict(filtered_albums)
            else:
                write(f"No match found", Color.RED)

    @staticmethod
    def _exit():
        sys.stdout.write('\033[?25h')
        sys.exit(0)



from enum import Enum
from dataclasses import dataclass


@dataclass(unsafe_hash=True)
class _Command:
    name: str
    description: str


class Command(Enum):
    SEARCH = _Command(
        name="search", 
        description=(
            "Searches for and provides music associated with the specified artist name.\n"
            "Some artists, although present, are not detected by the algorithm\n"
            "that is responsible for providing the information. In this case\n"
            "you need to consider using the 'id' command."
        )
    )

    ID = _Command(
        name="id", 
        description=(
            "Searches for and provides music associated with the specified ID.\n"
            "This command is useful when the artist of interest is not found\n"
            "by the 'search' command. The id can be extracted from the url of\n"
            "the artist's page (music.youtube.com).\n"
            "For example:\n"
            "https://music.youtube.com/channel/UCrpJvPlZprRX930AKTID1KA\n" 
            "                                  ^^^^^^^^^^^^^^^^^^^^^^^^"
        )
    )

    DOWNLOAD = _Command(
        name="download", 
        description=(
            "Starts downloading the specified media contents.\n"
            "The contents are specified through the respective indexes.\n"
            "If you intend to download only one content, simply provide\n"
            "its index as an argument to the command.\n"
            "Multiple contents can be specified as follows:\n"
            "\u2022 'all': download all contents.\n"
            "\u2022 Indices separated by a comma, for example 0, 1, ...\n"
            "\u2022 A slice that respects the Python syntax, for example\n"
            " 0:3 to download contents with index 0, 1, and 2."
        )
    )

    FILTER = _Command(
        name="filter",
        description=(
            "Filter the previously listed multimedia contents\n"
            "using the specified regular expression."
        )
    )

    EXIT = _Command(
        name="exit", 
        description="Exit the program."
    )

    HELP = _Command(
        name="help", 
        description="Print the help message."
    )

    CLEAR = _Command(
        name="clear", 
        description="Clear the terminal screen."
    )

    @classmethod
    def get_by_name(cls, name: str) -> "Command":
        return cls.__members__[name.replace("-", "_").upper()]



from typing import Dict, List, Tuple
from ctube.containers import Album


def extract_artist_id(data: Dict) -> str:
    return data["contents"]["tabbedSearchResultsRenderer"][
        "tabs"][0]["tabRenderer"]["content"]["sectionListRenderer"][
        "contents"][1]["musicCardShelfRenderer"]["title"]["runs"][0][
        "navigationEndpoint"]["browseEndpoint"]["browseId"]


def extract_albums(data: Dict) -> Tuple[List[Album], str]:
    albums: List[Album] = []
    for item in data["contents"]["singleColumnBrowseResultsRenderer"]["tabs"][0][
        "tabRenderer"]["content"]["sectionListRenderer"]["contents"][
            0]["gridRenderer"]["items"]:

        item_data = item["musicTwoRowItemRenderer"]
        album_type = item_data["subtitle"]["runs"][0]["text"]
        title = item_data["title"]["runs"][0]["text"]
        release_year = int(item_data["subtitle"]["runs"][-1]["text"])
        thumbnail_url = item_data["thumbnailRenderer"][
            "musicThumbnailRenderer"]["thumbnail"][
            "thumbnails"][-1]["url"]
        playlist_id = item_data["menu"]["menuRenderer"]["items"][
            0]["menuNavigationItemRenderer"]["navigationEndpoint"][
            "watchPlaylistEndpoint"]["playlistId"]
        albums.append(
            Album(
                title=title, 
                album_type=album_type,
                release_year=release_year,
                thumbnail_url=thumbnail_url,
                playlist_id=playlist_id
            )
        )
    return albums, data["header"][
        "musicHeaderRenderer"]["title"]["runs"][0]["text"]


def extract_search_suggestions(data: Dict) -> List[str]:
    search_suggestions_section: Dict = data["contents"][0][
        "searchSuggestionsSectionRenderer"]
    if "contents" in search_suggestions_section:
        contents = search_suggestions_section["contents"]
    else:
        contents: List[Dict] = search_suggestions_section["shelfDivider"]
    return [
        el["searchSuggestionRenderer"]["navigationEndpoint"][
            "searchEndpoint"]["query"]
        for el in contents
    ]



import os
from urllib.error import HTTPError, URLError
from pathvalidate import sanitize_filename
from http.client import IncompleteRead
from enum import Enum
from typing import Callable, Generator
from ctube.containers import Album, Song
from pytubefix import Playlist, Stream, YouTube
from pytubefix.exceptions import VideoUnavailable, RegexMatchError
from ctube.errors import NoMP4StreamAvailable, EmptyStreamQuery


class BaseURL(str, Enum):
    PLAYLIST = "https://music.youtube.com/playlist?list="


class Downloader:
    def __init__(
            self,
            output_path: str,
            on_complete_callback: Callable[[Song], None],
            on_progress_callback: Callable[[Song, int, int], None],
            skip_existing: bool = False,
            timeout: int = 5,
            max_retries: int = 2
    ):
        self.output_path = output_path
        self.on_complete_callback = on_complete_callback 
        self.on_progress_callback = on_progress_callback
        self.skip_existing = skip_existing
        self.timeout = timeout
        self.max_retries = max_retries

    @property
    def output_path(self) -> str:
        return self._output_path

    @output_path.setter
    def output_path(self, output_path: str) -> None:
        if not os.path.exists(output_path):
            os.makedirs(output_path, exist_ok=True)
        if not os.path.isdir(output_path):
            raise NotADirectoryError
        if not os.access(path=output_path, mode=os.W_OK):
            raise NotADirectoryError
        self._output_path = output_path

    def _on_complete_callback(self, data: Song, filepath: str) -> None:
        data.filepath = filepath
        self.on_complete_callback(data)

    def _on_progress_callback(self, data: Song, bytes_remaining: int, stream: Stream) -> None:
        filesize = stream.filesize
        bytes_received = filesize - bytes_remaining
        self.on_progress_callback(data, filesize, bytes_received)

    def download_album(
            self, 
            album: Album, 
            artist: str, 
            image_data: bytes, 
    ) -> Generator:
        output_path = os.path.join(
            os.path.join(
                self.output_path, sanitize_filename(artist)
            ), 
            sanitize_filename(album.title)
        )
        os.makedirs(output_path, exist_ok=True)

        playlist = Playlist(url=f"{BaseURL.PLAYLIST.value}{album.playlist_id}")
        for i, url in enumerate(playlist):
            youtube = YouTube(url=url)

            song = Song(
                title=youtube.title,
                artist=artist,
                track_num=i + 1,
                image_data=image_data,
                filepath="",
                album=album
            )

            youtube.register_on_progress_callback(
                lambda stream, _, bytes_remaining: self._on_progress_callback(
                    song, bytes_remaining, stream
                )
            )
            youtube.register_on_complete_callback(
                lambda _, filepath: self._on_complete_callback(
                    song, 
                    filepath  # type: ignore | another problem with pytubefix ?
                )
            )

            try:
                self._download_song(youtube, output_path=output_path)
            except (
                    VideoUnavailable,
                    IncompleteRead, 
                    TimeoutError,
                    EmptyStreamQuery,
                    NoMP4StreamAvailable,
                    HTTPError,
                    URLError,
                    RegexMatchError,
                    KeyError # https://github.com/JuanBindez/pytubefix/issues/88
            ) as err:
                # built-in exceptions do not have __module__ attr.
                if hasattr(err, "__module__"):
                    error = f"{err.__module__}.{err.__class__.__name__}: {err}"
                else:
                    error = f"{err.__class__.__name__}: {err}"
            else:
                error = None

            yield song, error

    def _download_song(self, youtube: YouTube, output_path: str) -> None:
        streams = youtube.streams
        if not len(streams):
            raise EmptyStreamQuery(f"The song '{youtube.title}' did not provide any data streams")
        else:
            stream = streams.get_audio_only(subtype="mp4")
            if stream is None:
                raise NoMP4StreamAvailable("Unexpected status: MP4 stream unavailable")

        stream.download(
            output_path=output_path,
            skip_existing=self.skip_existing,
            timeout=self.timeout,
            max_retries=self.max_retries
        )



from enum import Enum
from typing import Set


class Color(Enum):
    RESET = "\033[0m"
    BOLD = "\033[1m"
    RED = "\033[91m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    BLUE = "\033[94m"
    MAGENTA = "\033[95m"
    CYAN = "\033[96m"
    WHITE = "\033[97m"


def get_color_by_name(color_name: str) -> Color:
    return Color.__members__[color_name.upper()]


def color(string: str, col: Color) -> str:
    return f"{col.value}{string}{Color.RESET.value}"


def get_color_names() -> Set[str]:
    upper_color_names = set(Color.__members__.keys())
    upper_color_names.remove("RESET")
    return {color_name.lower() for color_name in upper_color_names}



from typing import Tuple, Optional, Union, List
from ctube.errors import InvalidIndexSyntax


def parse_user_input(user_input: str) -> Tuple[str, str]:
    splitted_user_input = user_input.split()
    prefix = splitted_user_input[0]
    arg = ' '.join(splitted_user_input[1:])
    return prefix, arg


def parse_indexes(string: str) -> Optional[Union[slice, List[int]]]:
    if string.lower() == "all":
        return None
    else:
        # x
        if string.isdigit():
            return slice(int(string), int(string) + 1)

        # x: y
        double_points_splitted_string = string.split(":")
        if len(double_points_splitted_string) > 1:
            if len(double_points_splitted_string) != 2:
                raise InvalidIndexSyntax("Missing stop index.")

            first, second = double_points_splitted_string
            if not first.isdigit() or not second.isdigit():
                raise InvalidIndexSyntax("Indexes must be integers.")

            first, second = int(first), int(second)
            if first < 0 or second < 0 or second <= first:
                raise InvalidIndexSyntax("Indexes must be positive integers.")

            return slice(first, second)

        # x, y, ..., z
        comma_splitted_string = string.split(",")
        if len(comma_splitted_string) > 1:
            try:
                indexes = list(map(int, comma_splitted_string))
            except ValueError:
                raise InvalidIndexSyntax("Indexes must be integers.")

            filtered_indexes: List[int] = []
            for index in indexes:
                if index < 0:
                    raise InvalidIndexSyntax("Indexes must be positive integers.")
                if index not in filtered_indexes:
                    filtered_indexes.append(index)
            return filtered_indexes
        raise InvalidIndexSyntax(f"Invalid index syntax: {string}")



class CTubeError(Exception):
    """ctube base exception. All exceptions contained in 
       this module inherit from this one."""


class InvalidIndexSyntax(CTubeError):
    """ Occurs when the user provides an argument 
        to the 'download' command with invalid syntax."""


class NoMP4StreamAvailable(CTubeError):
    """ Occurs when no stream with subtype mp4 is available. """


class EmptyStreamQuery(CTubeError):
    """This occurs when there are no streams for a song."""
